@@ -57,7 +57,7 @@ gdm_file_check (const gchar *caller, uid_t user, const gchar *dir,
 	    return FALSE;
 
     /* Stat directory */
-    IGNORE_EINTR (r = stat (dir, &statbuf));
+    VE_IGNORE_EINTR (r = stat (dir, &statbuf));
     if (r < 0) {
 	    if ( ! absentdirok)
 		 syslog (LOG_WARNING, _("%s: Directory %s does not exist."),
@@ -86,7 +86,7 @@ gdm_file_check (const gchar *caller, uid_t user, const gchar *dir,
     fullpath = g_build_filename (dir, file, NULL);
 
     /* Stat file */
-    IGNORE_EINTR (r = stat (fullpath, &statbuf));
+    VE_IGNORE_EINTR (r = stat (fullpath, &statbuf));
     if (r < 0) {
         /* Return true if file does not exist and that is ok */
 	if (absentok) { 
@@ -153,7 +153,7 @@ gdm_auth_file_check (const gchar *caller, uid_t user, const gchar *authfile, gbo
 	    return FALSE;
 
     /* Stat file */
-    IGNORE_EINTR (r = lstat (authfile, &statbuf));
+    VE_IGNORE_EINTR (r = lstat (authfile, &statbuf));
     if (s != NULL)
 	    *s = statbuf;
     if (r < 0) {
