@@ -26,6 +26,10 @@
 #include <sys/types.h>
 #include <unistd.h>
 
+#if defined(CAN_CLEAR_ADMCHG) && defined(HAVE_USERSEC_H)
+#  include <usersec.h>
+#endif /* CAN_CLEAR_ADMCHG && HAVE_USERSEC_H */
+
 #ifdef HAVE_CRYPT
 #  include <crypt.h>
 #endif /* HAVE_CRYPT */
@@ -42,6 +46,7 @@
 extern gboolean GdmAllowRoot;
 extern gboolean GdmAllowRemoteRoot;
 extern gint GdmRetryDelay;
+extern gboolean GdmDisplayLastLogin;
 
 static char *selected_user = NULL;
 
@@ -102,6 +107,14 @@ gdm_verify_user (GdmDisplay *d, const char *username, const gchar *display, gboo
     gchar *login, *passwd, *ppasswd;
     struct passwd *pwent;
     struct spwd *sp;
+#if defined(HAVE_PASSWDEXPIRED) && defined(HAVE_CHPASS) \
+    || defined(HAVE_LOGINRESTRICTIONS)
+    gchar *message = NULL;
+#endif
+#if defined(HAVE_PASSWDEXPIRED) && defined(HAVE_CHPASS)
+    gchar *info_msg = NULL, *response = NULL;
+    gint reEnter, ret;
+#endif
 
     if (local)
 	    gdm_slave_greeter_ctl_no_ret (GDM_STARTTIMER, "");
@@ -129,10 +142,10 @@ authenticate_again:
 	    }
 	    gdm_slave_greeter_ctl_no_ret (GDM_MSG, "");
 
-	    if ( ! gdm_slave_check_user_wants_to_log_in (login)) {
-		    g_free (login);
-		    login = NULL;
-		    goto authenticate_again;
+	    if (GdmDisplayLastLogin) {
+		    char *info = gdm_get_last_info (login);
+		    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, info);
+		    g_free (info);
 	    }
     } else {
 	    login = g_strdup (username);
@@ -226,6 +239,30 @@ authenticate_again:
 	    return NULL;
     }
 
+#ifdef HAVE_LOGINRESTRICTIONS
+
+    /* Check with the 'loginrestrictions' function
+       if the user has been disallowed */
+    if (loginrestrictions (login, 0, NULL, &message) != 0) {
+	    gdm_error (_("User %s not allowed to log in"), login);
+	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
+					  _("\nThe system administrator "
+					    "has disabled your "
+					    "account."));
+	    g_free (login);
+	    g_free (passwd);
+	    g_free (ppasswd);
+	    if (message != NULL)
+		    free (message);
+	    return NULL;
+    }
+    
+    if (message != NULL)
+	    free (message);
+    message = NULL;
+
+#else /* ! HAVE_LOGINRESTRICTIONS */
+
     /* check for the standard method of disallowing users */
     if (pwent->pw_shell != NULL &&
 	(strcmp (pwent->pw_shell, "/sbin/nologin") == 0 ||
@@ -243,9 +280,18 @@ authenticate_again:
 	    g_free (ppasswd);
 	    return NULL;
     }	
+
+#endif /* HAVE_LOGINRESTRICTIONS */
+
     g_free (passwd);
     g_free (ppasswd);
 
+    if ( ! gdm_slave_check_user_wants_to_log_in (login)) {
+	    g_free (login);
+	    login = NULL;
+	    goto authenticate_again;
+    }
+
     if ( ! gdm_setup_gids (login, pwent->pw_gid)) {
 	    gdm_error (_("Cannot set user group for %s"), login);
 	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
@@ -256,6 +302,108 @@ authenticate_again:
 	    return NULL;
     }
 
+#if defined(HAVE_PASSWDEXPIRED) && defined (HAVE_CHPASS)
+
+    switch (passwdexpired (login, &info_msg)) {
+    case 1 :
+	    gdm_error (_("Password of %s has expired"), login);
+	    gdm_error_box (d, GTK_MESSAGE_ERROR,
+			   _("You are required to change your password.\n"
+			     "Please choose a new one."));
+	    g_free (info_msg);
+
+	    do {
+		    ret = chpass (login, response, &reEnter, &message);
+		    g_free (response);
+
+		    if (ret != 1) {
+			    if (ret != 0) {
+				    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX,
+								  _("\nCannot change your password, "
+								    "you will not be able to log in, "
+								    "please try again later or contact "
+								    "your system administrator."));
+			    } else if ((reEnter != 0) && (message)) {
+				    response = gdm_slave_greeter_ctl (GDM_NOECHO, message);
+				    if (response == NULL)
+					    response = g_strdup ("");
+			    }
+		    }
+
+		    g_free(message);
+		    message = NULL;
+
+	    } while ( ((reEnter != 0) && (ret == 0))
+		      || (ret ==1) );
+
+	    g_free (response);
+	    g_free (message);
+
+	    if ((ret != 0) || (reEnter != 0)) {
+		    return NULL;
+	    }
+
+#if defined(CAN_CLEAR_ADMCHG)
+	    /* The password is changed by root, clear the ADM_CHG
+	       flag in the passwd file */
+	    ret = setpwdb (S_READ | S_WRITE);
+	    if (!ret) {
+		    upwd = getuserpw (login);
+		    if (upwd == NULL) {
+			    ret = -1;
+		    }
+		    else {
+			    upwd->upw_flags &= ~PW_ADMCHG;
+			    ret = putuserpw (upwd);
+			    if (!ret) {
+				    ret = endpwdb ();
+			    }
+		    }
+	    }
+
+	    if (ret) {
+		    gdm_error_box (d, GTK_MESSAGE_WARNING,
+				   _("Your password has been changed but "
+				     "you may have to change it again, "
+				     "please try again later or contact "
+				     "your system administrator."));
+	    }
+
+#else /* !CAN_CLEAR_ADMCHG */
+	    gdm_error_box (d, GTK_MESSAGE_WARNING,
+			   _("Your password has been changed but you "
+			     "may have to change it again, please try again "
+			     "later or contact your system administrator."));
+
+#endif /* CAN_CLEAR_ADMCHG */
+
+	    break;
+
+    case 2 :
+	    gdm_error (_("Password of %s has expired"), login);
+	    gdm_error_box (d, GTK_MESSAGE_ERROR,
+			   _("Your password has expired.\n"
+			     "Only a system administrator can now change it"));
+	    g_free (info_msg);
+	    return NULL;
+	    break;    
+
+    case -1 :
+	    gdm_error (_("Internal error on passwdexpired"));
+	    gdm_error_box (d, GTK_MESSAGE_ERROR,
+			   _("An internal error occured, you will not be able to log in.\n"
+			     "Please try again later or contact your system administrator."));
+	    g_free (info_msg);
+	    return NULL;
+	    break;    
+
+    default :
+	    g_free (info_msg);
+	    break;
+    }
+
+#endif /* HAVE_PASSWDEXPIRED && HAVE_CHPASS */
+
     return login;
 }
 
