@@ -32,9 +32,14 @@ greeter_item_info_new (GreeterItemInfo *parent,
   info->width_type = GREETER_ITEM_SIZE_UNSET;
   info->height_type = GREETER_ITEM_SIZE_UNSET;
 
-  for (i=0; i< GREETER_ITEM_STATE_MAX; i++)
+  if (type != GREETER_ITEM_TYPE_LIST)
     {
-      info->alphas[i] = 1.0;
+      for (i=0; i< GREETER_ITEM_STATE_MAX; i++)
+        {
+	  /* these happen to coincide for all
+	     items but list */
+          info->data.rect.alphas[i] = 0xff;
+        }
     }
 
   info->box_orientation = GTK_ORIENTATION_VERTICAL;
@@ -46,6 +51,13 @@ greeter_item_info_new (GreeterItemInfo *parent,
 
   info->button = FALSE;
 
+  if (GREETER_ITEM_TYPE_IS_TEXT (info))
+    {
+      info->data.text.max_width = 0xffff;
+      info->data.text.max_screen_percent_width = 90;
+      info->data.text.real_max_width = 0;
+    }
+
   return info;
 }
 
@@ -57,12 +69,18 @@ greeter_item_info_free (GreeterItemInfo *info)
 
   for (i = 0; i < GREETER_ITEM_STATE_MAX; i++)
     {
-      if (info->pixbufs[i])
-        g_object_unref (G_OBJECT (info->pixbufs[i]));
-      if (info->files[i])
-        g_free (info->files[i]);
-      if (info->fonts[i])
-        pango_font_description_free (info->fonts[i]);
+      if (GREETER_ITEM_TYPE_IS_PIXMAP (info))
+        {
+          if (info->data.pixmap.pixbufs[i] != NULL)
+            g_object_unref (G_OBJECT (info->data.pixmap.pixbufs[i]));
+          if (info->data.pixmap.files[i] != NULL)
+            g_free (info->data.pixmap.files[i]);
+	}
+      else if (GREETER_ITEM_TYPE_IS_TEXT (info))
+        {
+          if (info->data.text.fonts[i] != NULL)
+            pango_font_description_free (info->data.text.fonts[i]);
+	}
     }
 
   list = info->fixed_children;
@@ -75,8 +93,12 @@ greeter_item_info_free (GreeterItemInfo *info)
   g_list_foreach (list, (GFunc) greeter_item_info_free, NULL);
   g_list_free (list);
 
+  if (GREETER_ITEM_TYPE_IS_TEXT (info))
+    g_free (info->data.text.orig_text);
+
+  /* FIXME: what about custom list items! */
+
   g_free (info->id);
-  g_free (info->orig_text);
   g_free (info->show_type);
 
   memset (info, 0, sizeof (GreeterItemInfo));
@@ -88,9 +110,10 @@ greeter_item_update_text (GreeterItemInfo *info)
 {
   char *text;
   if (info && info->item &&
-      GNOME_IS_CANVAS_TEXT (info->item))
+      GNOME_IS_CANVAS_TEXT (info->item) &&
+      GREETER_ITEM_TYPE_IS_TEXT (info))
     {
-      text = greeter_item_expand_text (info->orig_text);
+      text = greeter_item_expand_text (info->data.text.orig_text);
 
       g_object_set (G_OBJECT (info->item),
 		    "markup", text,
@@ -117,7 +140,7 @@ get_clock (void)
     } 
   else 
     {
-      return ve_strftime (the_tm, _("%a %b %d, %I:%M %p"));
+      return ve_strftime (the_tm, _("%a %b %d, %l:%M %p"));
     }
 }
 
@@ -128,21 +151,53 @@ greeter_item_expand_text (const char *text)
   GString *str;
   const char *p;
   char *clock;
-  int r;
-  int underline = -1;
+  int r, i, n_chars;
+  gboolean underline = FALSE;
   char buf[256];
   struct utsname name;
 
   str = g_string_sized_new (strlen (text));
 
   p = text;
+  n_chars = g_utf8_strlen (text, -1);
+  i = 0;
   
-  while (*p)
+  while (i < n_chars)
     {
-      if (*p == '%')
+      gunichar ch;
+
+      ch = g_utf8_get_char (p);
+
+      /* Backslash commands */
+      if (ch == '\\')
+        {
+	  p = g_utf8_next_char (p);
+	  i++;
+	  ch = g_utf8_get_char (p);
+
+	  if (i >= n_chars || ch == '\0')
+	    {
+	      g_warning ("Unescaped \\ at end of text\n");
+	      goto bail;
+	    }
+	  else if (ch == 'n')
+	    g_string_append_unichar (str, '\n');
+	  else
+	    g_string_append_unichar (str, ch);
+	}
+      else if (ch == '%')
 	{
-	  p++;
-	  switch (*p)
+	  p = g_utf8_next_char (p);
+	  i++;
+	  ch = g_utf8_get_char (p);
+
+	  if (i >= n_chars || ch == '\0')
+	    {
+	      g_warning ("Unescaped %% at end of text\n");
+	      goto bail;
+	    }
+
+	  switch (ch)
 	    {
 	    case '%':
 	      g_string_append (str, "%");
@@ -179,35 +234,34 @@ greeter_item_expand_text (const char *text)
 	      g_string_append (str, clock);
 	      g_free (clock);
 	      break;
-	    case 0:
-	      g_warning ("Unescaped %% at end of text\n");
-	      goto bail;
-	      break;
 	    default:
-	      g_warning ("unknown escape code %%%c in text\n", *p);
+	      if (ch < 127)
+	        g_warning ("unknown escape code %%%c in text\n", (char)ch);
+	      else
+		g_warning ("unknown escape code %%(U%x) in text\n", (int)ch);
 	    }
 	}
-      else if (*p == '_')
+      else if (ch == '_')
         {
-	  underline = g_utf8_skip[*(guchar *)(p+1)];
+	  underline = TRUE;
 	  g_string_append (str, "<u>");
 	}
       else
 	{
-	  g_string_append_c (str, *p);
-	  if (underline >= 0)
+	  g_string_append_unichar (str, ch);
+	  if (underline)
 	    {
-	      underline--;
-	      if (underline == 0)
-	        g_string_append (str, "</u>");
+	      underline = FALSE;
+	      g_string_append (str, "</u>");
 	    }
 	}
-      p++;
+      p = g_utf8_next_char (p);
+      i++;
     }
   
  bail:
 
-  if (underline >= 0)
+  if (underline)
     g_string_append (str, "</u>");
 
   return g_string_free (str, FALSE);
